from setuptools import setup

setup(
    description=
        "The purpose of this Python namespace subpackage is to provide"
        " pure-python implementations of precise interval arithmetic.",
)
